#include <SoftwareSerial.h>
#include <string.h>
#include </home/davidc/Arduino/UFS4057ES_Attiny88_Firmware_.01b/delay_x.h>
#include <inttypes.h>
#define F_CPU 16000000UL // 16 MHz

SoftwareSerial mySerial(14,13); // RX, TX



//individual pin defs
const byte clkm   = 18;  // toggle clock mode (requires a pulse);
const byte cbe    = 17;  // enable the control bus output
const byte clk    = 8;   // clk pin
const byte hclk   = 9;   // half step the clock / set up a generated clock PWM;
const byte res    = 16;  // hard reset
const byte cmode  = 15;  // clock mode 

/*
 * Stack detals
 * 0 -     internal op register
 * 1 - 50  stack 
 * 51- 52  HL Register  
 * 53- 54  JMP Register
 * 55- 56  RTN Addres
 * 57- 59  UART 57 - Output next screen write, 58 - input next availble character or null,  59 - waiting character?
 *
 */

 /*Regeister Maps:  
      * 
      * each register is conntected to a 74HC138 demultiplexer / 74HC14 ST Inverter, the inputs to the demultiplexer are wired
      * to PORTC the lower 3 bits 0-3 control a trancever or tri-state buffer for output (except in the case of register 4).
      * the upper 3 bits 4-6 control the load functions of the registers.
      * 
      * it is important to note registers 4 and 7 are internal to the MCU, thier select lines are anded together and they share a 
      * transcever.  Therefore any transfer two and from is purely internal. there will be no bus activity when shareing data between
      * them.
      * 
      * Register input list: 
      * 
      *  1 - D Register input 
      *  2 - control registers input (alu and control unit)
      *  3 - writing any value to this Zeros the Flags.
      *  4 - Stack Register input
      *  5 - Lower 8 bits paralell load Adress Bus (not yet built)
      *  6 - Upper 8 bits paralell load Adress Bus (not yet built)
      *  7 - Set Stack Register Pointer
      * 
      * Register output list:
      * 
      *  1 - D Register output 
      *  2 - ALU Output
      *  3 - flags register output 
      *  4 - stack register output 
      *  5 - Lower 8 bits paralell load adress Bus (not yet built)
      *  6 - Upper 8 bits paralell load adress bus (not yet built)
      *  7 - Read Stack Register pointer
      *
      */
byte       tbase = 0xFF;   //minimum timebase single step.
byte       stack[59];   //50 byte stack
byte       stack_ptr = 1;    //position of the stack
bool       stack_of = false;

 
void setup()
{
  
  // defuault port states for ALM
  
  DDRD = B00000000; // port D == databus - input
  DDRC = B00111111; // port C == control bus - output nibble and a half
 
  
  pinMode(clkm, OUTPUT);
  pinMode(cbe, OUTPUT);
  pinMode(clk, INPUT);
  pinMode(hclk, OUTPUT);
  pinMode(res, OUTPUT);
  pinMode(cmode, INPUT); 
  
  // set the data rate for the SoftwareSerial port
  mySerial.begin(115200);
  /* mySerial.print("\033c");
     mySerial.println(" (\__/)  .~    ~.");
     mySerial.println(" /O O  ./      .'");
     mySerial.println("{O__,   \    {   ");
     mySerial.println("  / .  . )    \  ");
     mySerial.println("  |-| '-' \    } ");
     mySerial.println(" .(   _(   )_.'  ");
     mySerial.println("'---.~_ _ _&     ");
     mySerial.println(); */
     //mySerial.println("\a\033[2J\n\r__  (\\_\n\r(_ \\ ( '>\n\r  ) \\/_)=\n\r  (_(_ )_"); 
   
     
     mySerial.println("\a\033[2J\n\rUFS4057ES V.01b\n\r(C)2022 D.Collins");
  

}

void(* resetFunc) (void) = 0; // should soft reset, reset when called only resetting the attiny.


/**
 * hex2int
 * take a hex string and convert it to a 8bit number (max 2 hex digits)
 */

 /* thanks radhoo - stack overflow ... this saved my butt!                      
    https://stackoverflow.com/questions/10156409/convert-hex-string-char-to-int 
    simply droped the size of datatype to byte to save space from a 32 bit int  */
 
byte hex2int(char *hex) {
    byte val = 0;
    while (*hex) {
        // get current character then increment
        uint8_t byte = *hex++; 
        // transform hex character to the 4bit equivalent number, using the ascii table indexes
        if (byte >= '0' && byte <= '9') byte = byte - '0';
        else if (byte >= 'a' && byte <='f') byte = byte - 'a' + 10;
        else if (byte >= 'A' && byte <='F') byte = byte - 'A' + 10;    
        // shift 4 to make space for new digit, and add the 4 bits of the new digit 
        val = (val << 4) | (byte & 0xF);
    }
    return val;
}
bool con_yesno() {
  bool opbr = false;
  char c;
  mySerial.print("y/n? ");
  while(opbr != true) {
    if (mySerial.available() > 0) {
     c = toupper(mySerial.read());
     switch (c) {
      case 'Y':
       mySerial.write(c);
       opbr = true;
       return true;
       break;
      case 'N':
       mySerial.write(c);
       opbr = true;
       return false;
       break;
     }
    }
  }
}

byte con_gbyte(byte b_size) {                        // should get a specified size byte from the console
 char cbyte[b_size+1] = "";
 char c;
 bool opbr = false;
  
 while(opbr != true) {
  if (mySerial.available() > 0)  {
   c = toupper(mySerial.read());
   if (c == 3)  resetFunc();                        // ctrl-c soft resets preserving computer state but bringing back to the : prompt, without runing op.
   if (c == 13) {                                   // enter is pressed.
    opbr = true;
    c = 0;
    break;                                          // break out if we are done early, ex: 0-9 (number is 1 digit when max of 2 is specified)
   }
   else if (!(c >= 'A' && c <= 'F' || c >= '0' && c <= '9')) c = '0';
   mySerial.write(c);
   strncat(cbyte, &c, 1);
   if (strlen(cbyte) == b_size)  opbr = true;
 
  }
 }
 return hex2int(cbyte); // always evaluate as hex
}

void con_outop(byte c) {
  mySerial.write(c);
  if (c > 32) mySerial.write(' ');
}

void con_pntSD(void) {
 mySerial.println();
 mySerial.write('[');
 if (stack_ptr-1 < 0 ) mySerial.print("##"); else mySerial.print(stack[stack_ptr-1],HEX);
 mySerial.write(',');
 mySerial.print(stack[stack_ptr],HEX);
 mySerial.write(',');
 mySerial.print(stack[stack_ptr+1],HEX);
 mySerial.write(']');
 mySerial.write('(');
 mySerial.print(stack_ptr);
 mySerial.write(')'); 
}


void cnt_cm(void) {                                 // turn on/off the clock module oscillator
  PORTA = PORTA | B00000010;                        // pulse clock mode high
  PORTA = PORTA & B11111101;                        // pulse clock mode low
  mySerial.print((PINA & B00000100) >> 2);          // display clock state 0 = STOP, 1 = RUN
}

void cnt_cp(void) {
  if ((PINA & B00000100) >> 2 == 1) {               // if clock is in 'RUN'
    if ((PINB & 1) == 1 ) while ((PINB & 1) == 1);  // if CLK is HIGH wait till its low
    else {                                          // CLK is LOW so, wait until it is again LOW;
      while ((PINB & 1) == 0);
      while ((PINB & 1) == 1);
    }
  }
  else {                                            // Manually pulse the clock based on the time constant.
   for (int a=0; a<2; a++) {
   delay(tbase);
   PORTB = PORTB | B00000010;
   delay(100);
   PORTB = PORTB & B11111101;
   delay(tbase);
   }
  }
}

void cnt_ALU() {                                  // clock the ALU
  PORTA = PORTA | B00000001;                      // cwe pin high      
  //_delay_ns(120);
  cnt_cp();
  PORTA = PORTA & B11111110;                      // cwe pin low
}

void cnt_BUSE(byte op1,byte op2) {
  PORTC = (op1 << 3) | op2;
  _delay_ns(35);

}

void cnt_BUSD(void) {
  DDRD = B00000000;
  PORTC = 0;
  PORTD = 0;
}

void cpu_ST(byte op1, byte op2) {
  if (op1 == 4  || op1 == 7) {
     mcu_ST(op1,op2);                             // MCU Stores value in selected internal register   
  } 
  else {
   DDRD = B11111111;   
   PORTD = op2;                   
   cnt_BUSE(4, op1);                               // MCU writes to bus, selected register (op1) reads and stores value
   cnt_cp();
   cnt_BUSD();
  }
}

//static delay means the max clock speed is someplace near 28Mhz (well above what the MCU let alone the bread board)
void cpu_MV(byte op1, byte op2) {               //copy from register to register / program counter location
 if ( op2 == 4 || op2 == 7) DDRD = B00000000;   //reading into the mcu
 if ( op1 == 4 || op1 == 7) {                   //internal to the mcu
    DDRD = B11111111;
    PORTD = mcu_RD(op1);
 }
 cnt_BUSE(op1,op2);
 cnt_cp();                                      //pulse the clock to store external registers
 if ( op2 == 4 || op2 == 7) mcu_ST(op2,PIND);   //store to MCU internal register
 cnt_BUSD();
}

void mcu_ST(byte op1, byte op2) {
  switch (op1) {
    case 4:
     if (stack_ptr > 50 || stack_ptr == 0) {
      stack[stack_ptr] = op2;
     }
     else {
      stack[stack_ptr] = op2;                         //store value on top of stack register
      stack_ptr ++;                                   //stack inc
      if (stack_ptr == 51) { 
       stack_ptr = 1;
       stack_of = true;                              //stack overflow bit set!
       }
     }
     break;
     case 7:
      stack_ptr = op2;                              //change stack pointer to input
      break;
  }
}

byte mcu_RD(byte op1) {                             // read value of internal register
  byte ret = 0;
  switch (op1) {
   case 4: 
    if (stack_ptr > 50 || stack_ptr == 0) {         //if a static register
     ret = stack[stack_ptr]; 
    }
    else {
     ret = stack[stack_ptr];                         //return value on top
     stack[stack_ptr] = 0;                           //zero value at top of stack.
     if (stack_ptr != 1) stack_ptr --;               //stack dec stop at 1;  
    }
    break;
   case 7:
    ret = stack_ptr;                                 //return stack pointer
  }
  return(ret); 
}

/* Error checking
 *  --------------
 * 
 * op 1 -- checks for multiplexer realistic values 
 * op 2 -- checks for control register realistic values  
 * op 3 -- checks for stack pointer realistic values
 * op 4 -- checks to see of register values are equal
 * 
 * codes: 
 * 100  -- the entered register value is out of range
 * 101  -- the entered control code is aboove 31 0x1F
 * 102  -- the entered stack pointer register is out of range > 59 (0x39)
 * 103  -- the entered registers are equal and not valid.
 * 104  -- no test program exists
 */
bool mcu_error(byte op, byte oprnd1, byte oprnd2) {
 bool otp = false;
 switch (op) {
  case 1: 
    if (oprnd1 > 7 || oprnd1 < 1)  {                                          // value must be 1 - 7
      mySerial.println();
      mySerial.print("E:100");                                                  //100 The entered register value is out of range
      otp = true;
    }
  break;
  case 2:
    if (oprnd1 > 31) {                                                        //control register has max value of 31 (0x1F)
      mySerial.println();
      mySerial.print("E:101");                                                  //101 the entered control code is above 31 (0x1F)
      otp = true;  
    }
  break;
  case 3:
    if (oprnd1 > 59) {                                                        //stack pointer can not be > 59 (0x39)
      mySerial.println();
      mySerial.print("E:102");                                                  //102 the entered value for the stack pointer is higher than 59
      otp = true; 
    }
  break;
  case 4:
    if (oprnd1 == oprnd2) {                                                   //103 the operands can not be equal
      mySerial.println();
      mySerial.print("E:103");
      otp = true;      
    }
    break;
  case 5:
    mySerial.println();
    mySerial.print("E:104");
    otp = true;
    break;  
 }
 return otp;
}

/* test program for ALU 
 *  
 *  below is the ALU test program it is an XOR SHIFT. This test handles the jumps on the microcontroler because they 
 *  are not built on the main build. this is only to test the ALU data path.
 */

// add 2 16 bit numbers op1 is pointer to big end of operand 1, little end is op1+1 , op2 is same for operand2 
// uses HL for result.
// numbers are unsigned 16 bit
 
void tp_addFF(byte op1, byte op2) {
  cpu_ST(7,op1+1);        // set pointer to little end of op1
  cpu_MV(4,1);            // pop op1 little end to D
  cpu_ST(2, 0x0C);        // set ALU to Par. Load
  cnt_ALU();              // clock ALU
  cpu_ST(7,op2+1);        // set pointer to little end of op2
  cpu_MV(4,1);            // pop op2 little end to D, little end is ready for ADD OP
  cpu_ST(2, 0x06);        // set ALU to ADD
  cnt_ALU();              // Add little ends together stores result in A.
  cpu_ST(7,52);           // sets pointer to L
  cpu_MV(2,4);            // moves A to l, next we deal with flags.
  cpu_MV(3,1);            // moves F to D
  cpu_ST(2, 0x0C);        // set ALU to Par. load
  cnt_ALU();              // clock alu, flags in acumulator
  cpu_ST(1, 1);           // sets D to 00000001 (set mask for flags)
  cpu_ST(2, 1);           // sets alu to AND
  cnt_ALU();              // clocks alu A now contains just cary out flag
  cpu_ST(7,51);           // sets pointer to H
  cpu_MV(2, 4);           // moves A to H, now we add the high ends together
  cpu_ST(7,op1);          // pointer to big end of op1
  cpu_MV(4, 1);           // pop big end op1 to D
  cpu_ST(2, 0x0C);        // set ALU to paralell load
  cnt_ALU();              // clock ALU paralell load D  is in A
  cpu_ST(7, op2);         // set pointer to op2 big end
  cpu_MV(4, 1);           // pop big end op2 into D
  cpu_ST(2, 0x06);        // sets alu to ADD
  cnt_ALU();              // add big ends together store results in A, we need to add carry bit
  cpu_ST(7, 51);          // pointer to H ( we saved the carry bit here)
  cpu_MV(4, 1);           // move H to D
  cnt_ALU();              // adds carry bit if any (needs work, as this ignores the flags on the big end)
  cpu_MV(2, 4);           // move A to H  added number is stored in HL
}

void tp_setRP(byte op1) {
  cpu_ST(7,51);
  cpu_MV(4,1);
  cpu_ST(2,0x0c);
  cnt_ALU();
  cpu_ST(7,52);
  cpu_MV(4,1);
  cpu_ST(7,1);
  cpu_MV(2,4);
  cpu_MV(1,4);
  cpu_ST(4,0);
  cpu_ST(4,op1);
}

void loop() // run over and over 
{
 byte c;
 byte oprnd1 = 0;
 byte oprnd2 = 0;
 bool err = false;
 bool nextcm = false;
 bool brk = false;
 unsigned total = 0;

 con_pntSD();                                                                // get a command
 mySerial.write(':');
 while (nextcm != true) {
  if (mySerial.available() > 0)  {
   c = toupper(mySerial.read());
   switch (c) {
     case 13:
      con_outop(c);
      nextcm = true;
      break;
    
     case 3:
      resetFunc();                                                           // no fall out this is a soft reset!

    
     case 'R':                                                               // ST Operation on a direct load register, and pulse clock 1x (or wait for 1 pulse).
       con_outop(c);      
       oprnd1 = con_gbyte(1);
       mySerial.write(' ');
       oprnd2 = con_gbyte(2);
       err = mcu_error(1, oprnd1,0);
       if (oprnd1 == 2) err = mcu_error(2, oprnd2,0);
       if (oprnd1 == 7) err = mcu_error(3, oprnd2,0);
       if (err != true) cpu_ST(oprnd1,oprnd2);                                //manualy loads non-accumulator register
       nextcm = true;
     break;
      
     case 'T':                                                               //set the single pulse clockbase 1-255;
       con_outop(c);
       tbase = con_gbyte(2);        
       nextcm = true;
     break;
       
     case 'C':                                                               //Set the clock module to 'RUN'
       con_outop(c);
       cnt_cm();                                                             //trip the clock module into 'RUN' mode; (this changes the behavor of the cnt_cp funciton.)
       nextcm = true;
     break;

     case 'A':                                                               //Manual Pulse the ALU & Clock.
       con_outop(c);                                               
       cnt_ALU();                                                            // clock the ALU useing the A&D register & Control Word.
       nextcm = true;
     break;

     case 'M':                                                               //MV Operation moves from register to register
       con_outop(c);
       nextcm = true;
       oprnd1 = con_gbyte(1);
       mySerial.write(' ');
       oprnd2 = con_gbyte(1);
       err = mcu_error(1, oprnd1,0);
       err = mcu_error(1, oprnd2,0);
       err = mcu_error(4, oprnd1,oprnd2);
       if (err != true ) cpu_MV(oprnd1,oprnd2);                              //move oprnd1 to oprnd2
     break;
     
     case 'P':
       con_outop(c);
       nextcm = true;
       switch (con_gbyte(1)) {
       case 1: 
        mySerial.write(' ');
        oprnd1 = con_gbyte(2);
        mySerial.write(' ');
        oprnd2 = con_gbyte(2);
        mySerial.println();
        total = (stack[oprnd1] << 8) | stack[oprnd1+1];
        mySerial.print(total);
        mySerial.print(" + ");
        total = (stack[oprnd2] << 8) | stack[oprnd2+1];
        mySerial.println(total);
        tp_addFF(oprnd1,oprnd2);
        mySerial.println();
        mySerial.print("R: ");
        total = (stack[51] << 8) | stack[52];
        mySerial.println(total); 
       break;
       case 2:
        mySerial.write(' ');
        oprnd1 = con_gbyte(2);
        mySerial.println();
        while(brk != true) {
          if (mySerial.available() > 0)  {
            c = toupper(mySerial.read());
            if (c == 'c') brk = true;
          }
          tp_setRP(oprnd1);
          tp_addFF(1,3);
          total = (stack[51] << 8) | stack[52];
          mySerial.println(total);
        }
       break; 
      break;
       default:
        mcu_error(5,0,0);
       }
       
       
     
   
     break;
       
     case 26:                                                                //CTRL-Z hard reset
       con_outop;
       mySerial.print("Reset ");
       if (con_yesno()) PORTA = PORTA | B00001000;                           // this should hard reset the MCU, and all of the registers. 
      
       /* there is no fall out of the loop because pulling this pin high will send a low signal 
          to the AVR's reset pin! */
      
       
     default:
     mySerial.write('\a');
     }    
   }   
  }
} 
